SystemOrganization addCategory: #'FullText-Engines'!
SystemOrganization addCategory: #'FullText-Adaptors'!

!CharacterSet class methodsFor: '*fulltext' stamp: 'bmp 12/12/2004 21:00'!
allNumerical
	"return a set containing only alphabetic characters"
	^ self
		allMatching: [:char | char isDigit]! !

!CharacterSet class methodsFor: '*fulltext' stamp: 'bmp 12/12/2004 21:01'!
allSeperator
	"return a set containing only alphabetic characters"
	^ self
		allMatching: [:char | char isSeparator]! !

!CharacterSet class methodsFor: '*fulltext' stamp: 'bmp 12/12/2004 20:59'!
allAlphabetic
	"return a set containing only alphabetic characters"
	^self allMatching: [ :char | char isLetter].! !

!CharacterSet class methodsFor: '*fulltext' stamp: 'bmp 12/12/2004 21:00'!
allCharacters
	"return a set containing all characters"
	^self allMatching: [ :unused | true ]! !

!CharacterSet class methodsFor: '*fulltext' stamp: 'bmp 12/12/2004 20:58'!
allMatching: aBlock 
	"return a set containing all characters matching some predicate."
	| set character |
	set _ self empty.
	0
		to: 255
		do: [:ascii | 
			character _ Character value: ascii.
			(aBlock value: character)
				ifTrue: [set add: character]].
	^ set! !

Object subclass: #IndexAdaptor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'FullText-Adaptors'!

!IndexAdaptor commentStamp: 'sac 1/23/2002 19:45' prior: 0!
An adaptor is something that adapts a document for indexing. A document may be any entitity, a String, a Text, a descriptor of a network or filesystem entity.

Given a document, the adaptor returns search terms that occur in the document. It also canonicalizes search terms.

Thus, to make any document indexable, just write an adaptor for it. 
!

!IndexAdaptor methodsFor: 'as yet unclassified' stamp: 'sac 1/23/2002 19:46'!
canonicalize: aString
	"Given a string or a term, canonicalize it for searching. Canonicalization is used to remove unimportant differences (case, capitalization, etc) in search terms."
	self subclassResponsibility! !

!IndexAdaptor methodsFor: 'as yet unclassified' stamp: 'sac 1/23/2002 19:46'!
terms: aDocument
	"Given a document, give a Set of the search terms occuring in it."
	self subclassResponsibility! !

IndexAdaptor subclass: #DemoStringAdaptor
	instanceVariableNames: ''
	classVariableNames: 'Ignore CSAlphabetic CSNonAlphabetic'
	poolDictionaries: ''
	category: 'FullText-Adaptors'!

DemoStringAdaptor subclass: #DemoTextAdaptor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'FullText-Adaptors'!

!DemoTextAdaptor methodsFor: 'as yet unclassified' stamp: 'sac 1/22/2002 19:32'!
canonicalize: aString
	"Given a string or a term, canonicalize it for searching."
	^aString asLowercase
! !

!DemoTextAdaptor methodsFor: 'as yet unclassified' stamp: 'sac 1/22/2002 19:33'!
terms: aText
	"Given a document, give a list of the search terms occuring in it."
	| |
	^super terms: aText asString.! !

!DemoStringAdaptor methodsFor: 'as yet unclassified' stamp: 'sac 1/23/2002 17:18'!
terms: aString
	"Given a document, give a list of the search terms occuring in it."
	| words |

	words _ aString substrings: CSAlphabetic complement: CSNonAlphabetic.
	words _ words select: [ :word | (word size > 3) & (Ignore includes: word) not].
	^words
! !

!DemoStringAdaptor methodsFor: 'as yet unclassified' stamp: 'sac 1/22/2002 18:58'!
canonicalize: aString
	"Given a string or a term, canonicalize it for searching."
	^aString asLowercase
! !

!DemoStringAdaptor class methodsFor: 'as yet unclassified' stamp: 'sac 1/23/2002 18:22'!
initialize
	super initialize.
	CSAlphabetic _ CharacterSet allAlphabetic.
	CSNonAlphabetic _ CSAlphabetic complement.
	Ignore _ #(iftrue iffalse self value do aBlock) asSet.! !

Object subclass: #DocumentIndex
	instanceVariableNames: 'setClass indexAdaptor documentCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'FullText-Engines'!

!DocumentIndex commentStamp: 'sac 1/24/2002 18:41' prior: 0!
The base class for document indexes. This supports both allocation of new sets, and expensive search algorithms like substring matching.

	setClass is the class that is to be used when constructing results. Weak Sets have not been tested, but should also work. I support adding and removing documents, substring searching, and any-of matching.

This is an abstract class.!

!DocumentIndex methodsFor: 'searching' stamp: 'sac 1/24/2002 17:21'!
containsNoCopy: aTerm
	"Find all documents that contain aTerm. The resultant set is a clone of the internal one. Thus, it MUST NOT BE MUTATED. The returned value should not be fed into any union: or intersect: or other set-operation."
	^self dbase at: (indexAdaptor canonicalize: aTerm) ifAbsent: [^self setClass new]! !

!DocumentIndex methodsFor: 'feedback' stamp: 'sac 1/24/2002 18:37'!
allTerms
	"Return a set of all terms in the dbase"
	^self dbase keys.! !

DocumentIndex subclass: #ExactDocumentIndex
	instanceVariableNames: 'dbase'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'FullText-Engines'!

!ExactDocumentIndex commentStamp: '<historical>' prior: 0!
Implement a document index based on hash tables. 

Thus, this document index supports exact queries and expensive queries.!

!ExactDocumentIndex class methodsFor: 'as yet unclassified' stamp: 'sac 1/23/2002 15:35'!
withAdaptor: anIndexAdaptor withSet: aSetClass
	"Build an index over documents. IndexAdaptor is something responsible for interfacing documents to and from the index code, using aSetClass to store the individual sets."
	^self new initialize: anIndexAdaptor withSet: aSetClass.! !

!ExactDocumentIndex methodsFor: 'private' stamp: 'sac 1/23/2002 19:48'!
initialize: anIndexAdaptor withSet: aSetClass
	"See superclass method with same name."
	super initialize: anIndexAdaptor withSet: aSetClass.
	dbase _ Dictionary new.
! !

!ExactDocumentIndex methodsFor: 'private' stamp: 'sac 1/22/2002 19:55'!
dbase
	^dbase.
! !

!DocumentIndex methodsFor: 'private' stamp: 'sac 1/25/2002 01:15'!
initialize: anIndexAdaptor withSet: aSetClass
	"Initialize both the adaptor, and the type of Set that will be used internally."
	indexAdaptor _ anIndexAdaptor.
	setClass _ aSetClass.
	"To make sure that the class has been reinitialized."
	anIndexAdaptor class initialize.
	documentCount _ 0.
	! !

DocumentIndex subclass: #OrderedDocumentIndex
	instanceVariableNames: 'dbase'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'FullText-Engines'!

!OrderedDocumentIndex commentStamp: '<historical>' prior: 0!
Store search terms in skiplist. 

This supports ranged queries and prefix queries.!

!OrderedDocumentIndex methodsFor: 'searching' stamp: 'sac 1/24/2002 17:26'!
termRangeFrom: startString to: endString
	"Return the set of documents that contain any term in the range of [startString, endString]"
	| sum node startTerm endTerm |
	startTerm _ (indexAdaptor canonicalize: startString).
	endTerm _ (indexAdaptor canonicalize: endString).
	sum _ Set new.
	node _ dbase search: startTerm.
	node ifNil: [^sum]. "No key exists."
	[node key <= endTerm]
		whileTrue:  [sum add: node key.
					node _ node next.
					node ifNil: [^sum]].
	^sum! !

!OrderedDocumentIndex methodsFor: 'searching' stamp: 'sac 1/24/2002 17:27'!
termPrefixOf: aString
	"Return the set of terms that have a term that matches the prefix of the given string:"
	| sum node term |
	term _ (indexAdaptor canonicalize: aString).
	sum _ Set new.
	node _ dbase search: aString.
	node ifNil: [^sum]. "No key exists."
	[node key beginsWith: term]
		whileTrue:  [sum add: node key.
					node _ node next.
					node ifNil: [^sum]].
	^sum! !

!OrderedDocumentIndex class methodsFor: 'as yet unclassified' stamp: 'sac 1/23/2002 15:35'!
withAdaptor: anIndexAdaptor withSet: aSetClass
	"Build an index over documents. IndexAdaptor is something responsible for interfacing documents to and from the index code, using aSetClass to store the individual sets."
	^self new initialize: anIndexAdaptor withSet: aSetClass.! !

!OrderedDocumentIndex methodsFor: 'private' stamp: 'sac 1/22/2002 20:12'!
dbase
	^dbase.
! !

!OrderedDocumentIndex methodsFor: 'searching' stamp: 'sac 1/24/2002 17:25'!
prefixOf: aString
	"Return the set of documents that have a term that matches the prefix of the given string:"
	| sum node term |
	term _ (indexAdaptor canonicalize: aString).
	sum _ self setClass new.
	node _ dbase search: term.
	node ifNil: [^sum]. "No key exists."
	[node key beginsWith: term]
		whileTrue:  [sum union: node value.
					node _ node next.
					node ifNil: [^sum]].
	^sum! !

!OrderedDocumentIndex methodsFor: 'private' stamp: 'sac 1/23/2002 19:48'!
initialize: anIndexAdaptor withSet: aSetClass
	"See superclass method with same name."
	super initialize: anIndexAdaptor withSet: aSetClass.
	dbase _ SkipList new.
! !

!OrderedDocumentIndex methodsFor: 'searching' stamp: 'sac 1/24/2002 17:26'!
rangeFrom: startString to: endString
	"Return the set of documents that contain any term in the range of [startString, endString]"
	| sum node startTerm endTerm |
	startTerm _ (indexAdaptor canonicalize: startString).
	endTerm _ (indexAdaptor canonicalize: endString).
	sum _ self setClass new.
	node _ dbase search: startTerm.
	node ifNil: [^sum]. "No key exists."
	[node key <= endTerm]
		whileTrue:  [sum union: node value.
					node _ node next.
					node ifNil: [^sum]].
	^sum! !

!DocumentIndex methodsFor: 'updating' stamp: 'sac 1/25/2002 10:55'!
add: aDocument
	"Add a document into the database."
	^self add: aDocument withKey: aDocument! !

!DocumentIndex methodsFor: 'feedback' stamp: 'sac 1/25/2002 00:58'!
termsByFrequency
	"Return a SortedCollection of all terms, sorted by their frequency of occurance. Output is an ordered collection of associations of term and count."

	| sorted |
	sorted _ SortedCollection sortBlock: [:x :y | x value > y value].
	self dbase keysAndValuesDo: [ :key :val | 
             sorted add: (Association key: key value: val size)].
	^sorted

! !

!DocumentIndex methodsFor: 'searching' stamp: 'sac 1/24/2002 17:23'!
subStringOf: aString
	"Return all documents containing a substring of the requested string."

	| sum |

	sum _ self setClass new.
	self dbase associationsDo: [ :each | 
            (each key findString: (indexAdaptor canonicalize: aString)) = 0 
               ifFalse: [ sum union: each value] ].
	^sum 
		
		! !

!DocumentIndex methodsFor: 'searching' stamp: 'sac 1/24/2002 17:22'!
contains: aTerm
	"Find all documents that contain aTerm."
	^(self dbase at: (indexAdaptor canonicalize: aTerm) ifAbsent: [^self setClass new]) copy.! !

!DocumentIndex methodsFor: 'advanced updating' stamp: 'sac 1/25/2002 19:35'!
add: aDocument withKey: aKey
	"Add a document into the database under the given key. This is to make things that are stored by reference easier. When you search, you get the keys back. This method is to allow some pre-parsing of aKey when making the document. For simple indexing, use #add:"
	| words |
	words _ indexAdaptor terms: aDocument.
	words do: [ :word |
				|canonword set|
					"canonword _ (indexAdaptor canonicalize: word).
					set _ self dbase at: canonword ifAbsent: [setClass new]."
					set _ self getSetFor: word.
				     set add: aKey.
					"self dbase at: canonword put: set"].
	documentCount _ documentCount + 1.! !

!DocumentIndex methodsFor: 'searching' stamp: 'sac 1/28/2002 18:31'!
fastAllOf: aSet
	"Return the set of documents that matches all of the words in the given set. This version reorders the intersections for speed, but saves intermediate results."
	| sum |
	sum _ aSet collect: [:each | self containsNoCopy: each ].
	^sum intersectAllInto: self setClass.
! !

!DocumentIndex methodsFor: 'updating' stamp: 'sac 4/13/2002 07:58'!
condense
	"Condensing an index can make the index smaller (about 10-40%), But it disables all 
      future updates to the index. 

      Once condensed, you can neither add nor remove a document from the index. 
	 Furthermore, after being condensed, any search is allowed to return a non-Set Collection."

	"The operation it does, is it replaces each result set with an Array, and condenses the documents that are put into the array."
	self dbase associationsDo: [ :each | 
	      each value: each value asArray.].
	Smalltalk tenure "Insure the root table doesn't contain the dictionary."! !

!DocumentIndex methodsFor: 'searching' stamp: 'sac 1/25/2002 13:20'!
anyOf: aSet
	"Return the set of documents that matches any of the words in the given set."
	| sum |
	sum _ setClass new.
	"If the set is empty, return an empty set."
	"aSet size = 0 ifTrue: [^sum]."
	"If the set has one thing, return the set straight from the dbase without copying."
	"aSet size = 1 ifTrue: [^self dbase at: (indexAdaptor canonicalize: aSet anyOne) ifAbsent: [sum]]. DISABLED!!!! I wish to allow clients to destructively mutate the resultant set."

	"Otherwise, merge them together."	
	aSet do: [ :item |  
                  sum union: (self containsNoCopy: item)
           ].
	^sum
! !

!DocumentIndex methodsFor: 'updating' stamp: 'sac 1/28/2002 18:41'!
uncondense
	"Uncondensing a condensed index allows changes to be made, but makes the condensed index bigger. Note that if the adaptor does anything special during the condense, that operation will not be undone. For an index that has been modified heavily over a long period of time, uncondensing may restore some space. uncondensing rebuilds each set in the index."

	self dbase associationsDo: [ :each | |temp|
		  temp _ self setClass new. 
		  temp addAll: each value.
	       each value: temp].
	Smalltalk tenure. "Ensure the root table doesn't contain the dictionary."! !

!DocumentIndex methodsFor: 'updating' stamp: 'sac 1/25/2002 10:56'!
remove: aDocument
	"Remove a document into the database. Note that the document must be unchanged from when it was added in, to insure that it is correctly removed."
	^self remove: aDocument withKey: aDocument
! !

!DocumentIndex methodsFor: 'feedback' stamp: 'sac 1/25/2002 01:16'!
documentCount
	"Return a count of the number of documents indexed. Note that it is a little naive, but you shouldn't be adding documents more than once."
	^documentCount! !

!DocumentIndex methodsFor: 'updating' stamp: 'sac 1/25/2002 19:49'!
removeStopTerm: aWord
	"Remove a particular term from the database entirely. For example, removing the most frequent terms from a database as a stoplist check."
	| canonword |
	canonword _ indexAdaptor canonicalize: aWord.
	self dbase at: canonword put: setClass new.! !

!DocumentIndex methodsFor: 'private' stamp: 'sac 1/23/2002 16:12'!
setClass
	"Return the class of sets that this index uses."
	^setClass! !

!DocumentIndex methodsFor: 'searching' stamp: 'sac 1/24/2002 17:16'!
termSubStringOf: aString
	"Return all terms containing a substring of the requested string."

	| sum |

	sum _ Set new.
	self dbase associationsDo: [ :each | 
            (each key findString: (indexAdaptor canonicalize: aString)) = 0 
               ifFalse: [ sum add: each key] ].
	^sum 
		
		! !

!DocumentIndex methodsFor: 'private' stamp: 'sac 1/23/2002 15:10'!
dbase
	"Return the current database. It should be a collection that can be scanned/enumerated."
	self subclassResponsibility.! !

!DocumentIndex methodsFor: 'private' stamp: 'sac 1/25/2002 19:46'!
getSetFor: aWord
	"Get and return the set associated with a porticular (uncanonicalized) word. Automatically uncondense it if necessary."
	| canonword set |
	canonword _ indexAdaptor canonicalize: aWord.
	set _ self dbase at: canonword ifAbsent: [
               |x|   x _ setClass new. self dbase at: canonword put: x. ^x].
	
	set class == Array ifTrue: [set _ setClass new. 
                                    set addAll: set. 
                                    self dbase at: canonword put: set].
	^set
! !

!DocumentIndex methodsFor: 'advanced updating' stamp: 'sac 1/25/2002 19:29'!
remove: aDocument withKey: aKey
	"Remove a document into the database, where the document is indexed under the given key. Note that the document and key must be unchanged from when it was added in, to insure that it is correctly removed."
	| words |
	words _ indexAdaptor terms: aDocument.
	words do: [ :word |
				|set|
				"set _ self dbase at: (indexAdaptor canonicalize: word) ifAbsent: [setClass new]."
				 set _ self getSetFor: word.
                   set remove: aKey ifAbsent: []].
	documentCount _ documentCount - 1.
! !

!DocumentIndex methodsFor: 'searching' stamp: 'sac 1/25/2002 12:47'!
allOf: aSet
	"Return the set of documents that matches all of the words in the given set."
	| sum |
	aSet size = 0 ifTrue: [^setClass new].
	sum _ nil.
	aSet do: [ :each | 
				 sum ifNil: [sum _ setClass new addAll: (self containsNoCopy: each)]
                          ifNotNil: [sum _ sum intersection: (self containsNoCopy: each)]].
	^sum.! !

!String methodsFor: '*fulltext' stamp: 'bmp 12/12/2004 21:53'!
= aString 
	"Answer whether the receiver sorts equally as aString.
	The collation order is simple ascii (with case differences)."

	aString species == String ifFalse: [
		aString isText ifTrue: [^ self = aString string].
		^ false].
	
	self == aString ifTrue: [ ^ true ].
	^ (self compare: self with: aString collated: AsciiOrder) = 2! !

!String methodsFor: '*fulltext' stamp: 'bmp 12/12/2004 21:48'!
substrings: aCharacterSet complement: aCharacterSetComplement
        "Answer an array of the substrings containing only charactes in aCharacterSet that compose the receiver, the other characterset should be the complement of the first. This is an optimization to avoid repeated complementing of sets."
        | result end beginning seperators |

        result _ WriteStream on: (Array new: 10).
        seperators _ aCharacterSetComplement.


        end _ 0.
        "find one substring each time through this loop"
        [ 
                "find the beginning of the next substring"
                beginning _ self indexOfAnyOf: aCharacterSet startingAt: end+1 ifAbsent: [ nil ].
                beginning ~~ nil ] 
        whileTrue: [
                "find the end"
                end _ self indexOfAnyOf: seperators startingAt: beginning ifAbsent: [ self size + 1 ].
                end _ end - 1.

                result nextPut: (self copyFrom: beginning to: end).

        ].


        ^result contents! !
DemoStringAdaptor initialize!
